# Generated from JSON.g4 by ANTLR 4.7.2
from antlr4 import *  # @UnusedWildImport

if __name__ is not None and "." in __name__:
    from .JSONParser import JSONParser


# else:
#     from JSONParser import JSONParser

# This class defines a complete listener for a parse tree produced by JSONParser.
class JSONListener(ParseTreeListener):

    # Enter a parse tree produced by JSONParser#json.
    def enterJson(self, ctx: JSONParser.JsonContext):
        pass

    # Exit a parse tree produced by JSONParser#json.
    def exitJson(self, ctx: JSONParser.JsonContext):
        pass

    # Enter a parse tree produced by JSONParser#envelope.
    def enterEnvelope(self, ctx: JSONParser.EnvelopeContext):
        pass

    # Exit a parse tree produced by JSONParser#envelope.
    def exitEnvelope(self, ctx: JSONParser.EnvelopeContext):
        pass

    # Enter a parse tree produced by JSONParser#obj.
    def enterObj(self, ctx: JSONParser.ObjContext):
        pass

    # Exit a parse tree produced by JSONParser#obj.
    def exitObj(self, ctx: JSONParser.ObjContext):
        pass

    # Enter a parse tree produced by JSONParser#type_pair.
    def enterType_pair(self, ctx: JSONParser.Type_pairContext):
        pass

    # Exit a parse tree produced by JSONParser#type_pair.
    def exitType_pair(self, ctx: JSONParser.Type_pairContext):
        pass

    # Enter a parse tree produced by JSONParser#qualifiedName.
    def enterQualifiedName(self, ctx: JSONParser.QualifiedNameContext):
        pass

    # Exit a parse tree produced by JSONParser#qualifiedName.
    def exitQualifiedName(self, ctx: JSONParser.QualifiedNameContext):
        pass

    # Enter a parse tree produced by JSONParser#pair.
    def enterPair(self, ctx: JSONParser.PairContext):
        pass

    # Exit a parse tree produced by JSONParser#pair.
    def exitPair(self, ctx: JSONParser.PairContext):
        pass

    # Enter a parse tree produced by JSONParser#propname.
    def enterPropname(self, ctx: JSONParser.PropnameContext):
        pass

    # Exit a parse tree produced by JSONParser#propname.
    def exitPropname(self, ctx: JSONParser.PropnameContext):
        pass

    # Enter a parse tree produced by JSONParser#array.
    def enterArray(self, ctx: JSONParser.ArrayContext):
        pass

    # Exit a parse tree produced by JSONParser#array.
    def exitArray(self, ctx: JSONParser.ArrayContext):
        pass

    # Enter a parse tree produced by JSONParser#value.
    def enterValue(self, ctx: JSONParser.ValueContext):
        pass

    # Exit a parse tree produced by JSONParser#value.
    def exitValue(self, ctx: JSONParser.ValueContext):
        pass




class AnsonException:
    type = "io.odysz.ansons.x.AnsonException"
    excode = 0
    err = ""

    def __init__(self, excode: int, template: str, *param: object):
        super().__init__()
        self.excode = excode
        self.err = template if param is None else template.format(param)



import unittest
from typing import cast

from src.anson.io.odysz.ansons import Anson
from test.io.oz.jserv.docs.syn.singleton import AppSettings
from test.io.oz.syn import SynodeConfig, AnRegistry, Synode, SynOrg


class YellowPagesTests(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None

    def testAnregistry(self):
        Anson.java_src('test')

        settings: AppSettings = cast(AppSettings, Anson.from_file('test/json/registry/settings.json'))

        self.assertEqual(type(settings), AppSettings)
        self.assertEqual('http://192.168.0.0:8964/jserv-album', settings.jservs['X'])
        self.assertEqual('''{
  "type": "io.oz.jserv.docs.syn.singleton.AppSettings",
  "envars": {"WEBROOT_Album-web": "example.com"},
  "vol_name": "VOLUME_HOME",
  "volume": "~/github/semantic-jserv/jserv-album/vol",
  "port": 8964,
  "jservs": {
    "X": "http://192.168.0.0:8964/jserv-album",
    "Y": "http://192.168.0.0:8965/jserv-album"},
  "installkey": null,
  "rootkey": "0123456789ABCDEF"
}''', settings.toBlock(beautify=True))

        diction: AnRegistry = cast(AnRegistry, Anson.from_file('test/json/registry/dictionary.json'))
        self.assertEqual(AnRegistry, type(diction))
        self.assertEqual(SynodeConfig, type(diction.config))
        self.assertEqual(SynOrg, type(diction.config.org))
        self.assertEqual(list, type(diction.config.peers))
        self.assertEqual(2, len(diction.config.peers))
        self.assertEqual(Synode, type(diction.config.peers[0]))
        self.assertEqual(0, diction.config.peers[0].nyq)
        # print(diction.toBlock())
        self.assertEqual('''{
  "type": "io.oz.syn.AnRegistry",
  "config": {
    "type": "io.oz.syn.SynodeConfig",
    "https": false,
    "synid": "SAMPLE-X",
    "domain": "zsu",
    "admin": "ody",
    "sysconn": "sys",
    "synconn": "docsyn",
    "org": {
      "type": "io.oz.syn.SynOrg",
      "meta": "io.oz.jserv.docs.meta.DocOrgMeta",
      "orgId": "ura",
      "orgName": "URA",
      "orgType": "nation",
      "market": "test",
      "webroot": "$WEBROOT_Album-web-in-settings.json",
      "homepage": "",
      "album0": "album-test"
    },
    "syncIns": 20,
    "peers": [{
      "type": "io.odysz.semantic.syn.Synode",
      "synid": "X",
      "org": "ura",
      "mac": "#URA.X",
      "nyq": 0,
      "domain": "zsu",
      "syn_uid": "X,X"
    }, {
      "type": "io.odysz.semantic.syn.Synode",
      "synid": "Y",
      "org": "ura",
      "mac": "#URA.Y",
      "nyq": 0,
      "domain": "zsu",
      "syn_uid": "Y,Y"
    }]
  },
  "synusers": [{
    "type": "io.odysz.semantic.syn.SyncUser",
    "userId": "ody",
    "userName": null,
    "pswd": "8964",
    "iv": null,
    "org": "ura",
    "domain": "zsu"
  }, {
    "type": "io.odysz.semantic.syn.SyncUser",
    "userId": "syrskyi",
    "userName": null,
    "pswd": "слава україні",
    "iv": null,
    "org": "ura",
    "domain": "zsu"
  }]
}''', diction.toBlock(beautify=True))


if __name__ == '__main__':
    unittest.main()
    t = YellowPagesTests()
    t.testAnregistry()




import dataclasses
import os
import unittest

from src.anson.io.odysz.common import Utils

"""
    PyCharm Debug Configuration:
    run as module: test.test
    working folder: py3
"""

def run_script(script_path):
    python = 'py' if Utils.iswindows() else 'python3'
    os.system(f'{python} {script_path}')

test_loader = unittest.TestLoader()
# test_suite = test_loader.discover(start_dir='.', pattern='test_*.py')
# test_suite = test_loader.discover(start_dir='.', pattern='test_00_Anson.py')
# test_suite = test_loader.discover(start_dir='.', pattern='test_01_Instance4Name.py')
# test_suite = test_loader.discover(start_dir='.', pattern='test_02_Ping.py')
test_suite = test_loader.discover(start_dir='.', pattern='test_*.py')

# Run the tests
def tryModulename():
    module_name = 'test.io.odysz.semantic.jprotocol'
    class_name = 'AnsonBody'

    # Dynamically import the module
    module = __import__(module_name, fromlist=[class_name])

    # Get the class
    cls = getattr(module, class_name)

    # Create an instance
    obj = cls()  # Adjust arguments as needed
    cls.__type__=f'{module_name}.{class_name}'
    # obj.uri = 'uri'

    _FIELDS = '__dataclass_fields__'  # see dataclasses.fields()
    fds = getattr(cls, _FIELDS)
    missingAttrs = []
    for k in fds:
        if not hasattr(obj, k) and fds[k].default is dataclasses.MISSING: #_MISSING_TYPE:
            missingAttrs.append(k)
            setattr(obj, k, None)

    if len(missingAttrs) > 0:
        Utils.warn(f'Missing attributes in {module_name}.{class_name}: {missingAttrs}. Anson expect a __init__() for all initialize the none default fields.')

    print(obj)

if __name__ == '__main__':

    # tryModulename()
    unittest.TextTestRunner(verbosity=2).run(test_suite)


'''
A temporary solution without LL* parser.

For testing another cheap way of deserialize JSON in python, without Antlr4,
unless the stream mode is critical.

- Semantics consistence with Java is to be verified.

- No need for generating Python3 source code from JSON ?

'''
import unittest
from dataclasses import dataclass
from typing import Any

from src.anson.io.odysz.ansons import Anson, _fields
from test.testier.extra import ExtraData


# https://colab.research.google.com/drive/1pqeZGfqdEl_kOlJQ76SCeuKTtD3NGlev


@dataclass
class MyDataClass(Anson):
    name: str
    age: int
    incumbent: bool
    extra: ExtraData
    items: list[Any]  # = field(default_factory=list)

    def __init__(self, name: str = '', age: int = '0'):
        super().__init__()
        self.extra = ExtraData()
        self.name = name
        self.age = age
        self.incumbent = False
        self.items = ['']  # field(default_factory=list)

class DataClassTest(unittest.TestCase):
    def testStr(self):
        foo = MyDataClass('Trump', 78)
        foo.extra.l = ['']
        # print(f'{foo.extra.__module__}.{foo.__class__.__name__}')
        self.assertEqual('test.testier.extra.MyDataClass',
                         f'{foo.extra.__module__}.{foo.__class__.__name__}')

        # __type__: Anson.Trumpfield(name='__type__', fieldtype=<class 'str'>, ...)
        # name: 'name': Anson.Trumpfield(name='name', ...)

        # self.assertEqual(_fields(foo, None)['age'].type, int)

        my = MyDataClass('zz', 12)
        my.incumbent = True
        mytype = type(my)
        print(my.toBlock())
        self.assertEqual(MyDataClass, mytype)
        self.assertEqual('''{
  "type": "test.test_00_Anson.MyDataClass",
  "extra": {
    "type": "test.testier.extra.ExtraData",
    "s": null,
    "i": 0,
    "l": null,
    "d": null
  },
  "name": "zz",
  "age": 12,
  "incumbent": true,
  "items": [""]
}''', my.toBlock(beautify=True))

        your = mytype('yy', 13)
        # print(your.toBlock())
        self.assertEqual('''{
  "type": "test.test_00_Anson.MyDataClass",
  "extra": {
    "type": "test.testier.extra.ExtraData",
    "s": null,
    "i": 0,
    "l": null,
    "d": null
  },
  "name": "yy",
  "age": 13,
  "incumbent": false,
  "items": [""]
}''', your.toBlock(beautify=True))

        jsonstr = '{"type": "test.test_00_Anson.MyDataClass", "name": "Trump", "age": 78, "incumbent": true, "extra": {"s": "sss", "i": 1, "l": 2, "d": {"u": "uuu"}}}'
        his = Anson.from_json(jsonstr)
        # print(his.name)
        self.assertEqual('Trump', his.name)
        self.assertTrue(his.incumbent)
        print(his)

        jsonstr = '{\
          "type": "test.test_00_Anson.MyDataClass",\
          "extra": {\
            "s": null,\
            "i": 0,\
            "l": ["a", 2],\
            "d": {}\
          },\
          "name": "zz",\
          "age": 12,\
          "items": ['']\
        }'
        her = Anson.from_json(jsonstr)
        # print(her.name, type(her))
        self.assertEqual('zz', her.name)
        self.assertEqual(MyDataClass, type(her))
        self.assertFalse(her.incumbent)
        # print(her.toBlock())
        self.assertEqual('''{
  "type": "test.test_00_Anson.MyDataClass",
  "extra": {
    "type": "test.testier.extra.ExtraData",
    "s": null,
    "i": 0,
    "l": ["a", 2],
    "d": {}
  },
  "name": "zz",
  "age": 12,
  "incumbent": false,
  "items": []
}''', her.toBlock(beautify=True))


if __name__ == '__main__':
    unittest.main()
    t = DataClassTest()
    t.testStr()



# Generated from JSON.g4 by ANTLR 4.7.2
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys


def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\23")
        buf.write("i\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7\4\b")
        buf.write("\t\b\4\t\t\t\4\n\t\n\3\2\3\2\3\2\7\2\30\n\2\f\2\16\2\33")
        buf.write("\13\2\3\3\3\3\3\3\3\3\7\3!\n\3\f\3\16\3$\13\3\3\3\3\3")
        buf.write("\3\4\3\4\3\4\3\4\7\4,\n\4\f\4\16\4/\13\4\3\4\3\4\3\4\3")
        buf.write("\4\5\4\65\n\4\3\5\3\5\3\5\3\5\3\6\3\6\5\6=\n\6\3\6\3\6")
        buf.write("\3\6\7\6B\n\6\f\6\16\6E\13\6\3\6\5\6H\n\6\3\7\3\7\3\7")
        buf.write("\3\7\3\b\3\b\3\t\3\t\3\t\3\t\7\tT\n\t\f\t\16\tW\13\t\3")
        buf.write("\t\3\t\3\t\3\t\5\t]\n\t\3\n\3\n\3\n\3\n\3\n\3\n\3\n\3")
        buf.write("\n\5\ng\n\n\3\n\2\2\13\2\4\6\b\n\f\16\20\22\2\4\3\2\t")
        buf.write("\n\3\2\20\21\2o\2\24\3\2\2\2\4\34\3\2\2\2\6\64\3\2\2\2")
        buf.write("\b\66\3\2\2\2\nG\3\2\2\2\fI\3\2\2\2\16M\3\2\2\2\20\\\3")
        buf.write("\2\2\2\22f\3\2\2\2\24\31\5\4\3\2\25\26\7\3\2\2\26\30\5")
        buf.write("\4\3\2\27\25\3\2\2\2\30\33\3\2\2\2\31\27\3\2\2\2\31\32")
        buf.write("\3\2\2\2\32\3\3\2\2\2\33\31\3\2\2\2\34\35\7\4\2\2\35\"")
        buf.write("\5\b\5\2\36\37\7\3\2\2\37!\5\f\7\2 \36\3\2\2\2!$\3\2\2")
        buf.write("\2\" \3\2\2\2\"#\3\2\2\2#%\3\2\2\2$\"\3\2\2\2%&\7\5\2")
        buf.write("\2&\5\3\2\2\2\'(\7\4\2\2(-\5\f\7\2)*\7\3\2\2*,\5\f\7\2")
        buf.write("+)\3\2\2\2,/\3\2\2\2-+\3\2\2\2-.\3\2\2\2.\60\3\2\2\2/")
        buf.write("-\3\2\2\2\60\61\7\5\2\2\61\65\3\2\2\2\62\63\7\4\2\2\63")
        buf.write("\65\7\5\2\2\64\'\3\2\2\2\64\62\3\2\2\2\65\7\3\2\2\2\66")
        buf.write("\67\7\17\2\2\678\7\6\2\289\5\n\6\29\t\3\2\2\2:;\7\7\2")
        buf.write("\2;=\7\b\2\2<:\3\2\2\2<=\3\2\2\2=>\3\2\2\2>C\7\20\2\2")
        buf.write("?@\t\2\2\2@B\7\20\2\2A?\3\2\2\2BE\3\2\2\2CA\3\2\2\2CD")
        buf.write("\3\2\2\2DH\3\2\2\2EC\3\2\2\2FH\7\21\2\2G<\3\2\2\2GF\3")
        buf.write("\2\2\2H\13\3\2\2\2IJ\5\16\b\2JK\7\6\2\2KL\5\22\n\2L\r")
        buf.write("\3\2\2\2MN\t\3\2\2N\17\3\2\2\2OP\7\7\2\2PU\5\22\n\2QR")
        buf.write("\7\3\2\2RT\5\22\n\2SQ\3\2\2\2TW\3\2\2\2US\3\2\2\2UV\3")
        buf.write("\2\2\2VX\3\2\2\2WU\3\2\2\2XY\7\13\2\2Y]\3\2\2\2Z[\7\7")
        buf.write("\2\2[]\7\13\2\2\\O\3\2\2\2\\Z\3\2\2\2]\21\3\2\2\2^g\7")
        buf.write("\21\2\2_g\7\22\2\2`g\5\6\4\2ag\5\4\3\2bg\5\20\t\2cg\7")
        buf.write("\f\2\2dg\7\r\2\2eg\7\16\2\2f^\3\2\2\2f_\3\2\2\2f`\3\2")
        buf.write("\2\2fa\3\2\2\2fb\3\2\2\2fc\3\2\2\2fd\3\2\2\2fe\3\2\2\2")
        buf.write("g\23\3\2\2\2\f\31\"-\64<CGU\\f")
        return buf.getvalue()


class JSONParser ( Parser ):

    grammarFileName = "JSON.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "','", "'{'", "'}'", "':'", "'['", "'L'", 
                     "'.'", "'$'", "']'", "'true'", "'false'", "'null'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "TYPE", "IDENTIFIER", "STRING", "NUMBER", 
                      "WS" ]

    RULE_json = 0
    RULE_envelope = 1
    RULE_obj = 2
    RULE_type_pair = 3
    RULE_qualifiedName = 4
    RULE_pair = 5
    RULE_propname = 6
    RULE_array = 7
    RULE_value = 8

    ruleNames =  [ "json", "envelope", "obj", "type_pair", "qualifiedName", 
                   "pair", "propname", "array", "value" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    TYPE=13
    IDENTIFIER=14
    STRING=15
    NUMBER=16
    WS=17

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7.2")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None




    class JsonContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def envelope(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JSONParser.EnvelopeContext)
            else:
                return self.getTypedRuleContext(JSONParser.EnvelopeContext,i)


        def getRuleIndex(self):
            return JSONParser.RULE_json

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterJson" ):
                listener.enterJson(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitJson" ):
                listener.exitJson(self)




    def json(self):

        localctx = JSONParser.JsonContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_json)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 18
            self.envelope()
            self.state = 23
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==JSONParser.T__0:
                self.state = 19
                self.match(JSONParser.T__0)
                self.state = 20
                self.envelope()
                self.state = 25
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class EnvelopeContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def type_pair(self):
            return self.getTypedRuleContext(JSONParser.Type_pairContext,0)


        def pair(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JSONParser.PairContext)
            else:
                return self.getTypedRuleContext(JSONParser.PairContext,i)


        def getRuleIndex(self):
            return JSONParser.RULE_envelope

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterEnvelope" ):
                listener.enterEnvelope(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitEnvelope" ):
                listener.exitEnvelope(self)




    def envelope(self):

        localctx = JSONParser.EnvelopeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_envelope)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 26
            self.match(JSONParser.T__1)
            self.state = 27
            self.type_pair()
            self.state = 32
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==JSONParser.T__0:
                self.state = 28
                self.match(JSONParser.T__0)
                self.state = 29
                self.pair()
                self.state = 34
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 35
            self.match(JSONParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ObjContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pair(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JSONParser.PairContext)
            else:
                return self.getTypedRuleContext(JSONParser.PairContext,i)


        def getRuleIndex(self):
            return JSONParser.RULE_obj

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterObj" ):
                listener.enterObj(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitObj" ):
                listener.exitObj(self)




    def obj(self):

        localctx = JSONParser.ObjContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_obj)
        self._la = 0 # Token type
        try:
            self.state = 50
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,3,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 37
                self.match(JSONParser.T__1)
                self.state = 38
                self.pair()
                self.state = 43
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==JSONParser.T__0:
                    self.state = 39
                    self.match(JSONParser.T__0)
                    self.state = 40
                    self.pair()
                    self.state = 45
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 46
                self.match(JSONParser.T__2)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 48
                self.match(JSONParser.T__1)
                self.state = 49
                self.match(JSONParser.T__2)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class Type_pairContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TYPE(self):
            return self.getToken(JSONParser.TYPE, 0)

        def qualifiedName(self):
            return self.getTypedRuleContext(JSONParser.QualifiedNameContext,0)


        def getRuleIndex(self):
            return JSONParser.RULE_type_pair

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterType_pair" ):
                listener.enterType_pair(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitType_pair" ):
                listener.exitType_pair(self)




    def type_pair(self):

        localctx = JSONParser.Type_pairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_type_pair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 52
            self.match(JSONParser.TYPE)
            self.state = 53
            self.match(JSONParser.T__3)
            self.state = 54
            self.qualifiedName()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class QualifiedNameContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self, i:int=None):
            if i is None:
                return self.getTokens(JSONParser.IDENTIFIER)
            else:
                return self.getToken(JSONParser.IDENTIFIER, i)

        def STRING(self):
            return self.getToken(JSONParser.STRING, 0)

        def getRuleIndex(self):
            return JSONParser.RULE_qualifiedName

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterQualifiedName" ):
                listener.enterQualifiedName(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitQualifiedName" ):
                listener.exitQualifiedName(self)




    def qualifiedName(self):

        localctx = JSONParser.QualifiedNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_qualifiedName)
        self._la = 0 # Token type
        try:
            self.state = 69
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [JSONParser.T__4, JSONParser.IDENTIFIER]:
                self.enterOuterAlt(localctx, 1)
                self.state = 58
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==JSONParser.T__4:
                    self.state = 56
                    self.match(JSONParser.T__4)
                    self.state = 57
                    self.match(JSONParser.T__5)


                self.state = 60
                self.match(JSONParser.IDENTIFIER)
                self.state = 65
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==JSONParser.T__6 or _la==JSONParser.T__7:
                    self.state = 61
                    _la = self._input.LA(1)
                    if not(_la==JSONParser.T__6 or _la==JSONParser.T__7):
                        self._errHandler.recoverInline(self)
                    else:
                        self._errHandler.reportMatch(self)
                        self.consume()
                    self.state = 62
                    self.match(JSONParser.IDENTIFIER)
                    self.state = 67
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass
            elif token in [JSONParser.STRING]:
                self.enterOuterAlt(localctx, 2)
                self.state = 68
                self.match(JSONParser.STRING)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PairContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def propname(self):
            return self.getTypedRuleContext(JSONParser.PropnameContext,0)


        def value(self):
            return self.getTypedRuleContext(JSONParser.ValueContext,0)


        def getRuleIndex(self):
            return JSONParser.RULE_pair

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPair" ):
                listener.enterPair(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPair" ):
                listener.exitPair(self)




    def pair(self):

        localctx = JSONParser.PairContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_pair)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 71
            self.propname()
            self.state = 72
            self.match(JSONParser.T__3)
            self.state = 73
            self.value()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class PropnameContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(JSONParser.STRING, 0)

        def IDENTIFIER(self):
            return self.getToken(JSONParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return JSONParser.RULE_propname

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterPropname" ):
                listener.enterPropname(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitPropname" ):
                listener.exitPropname(self)




    def propname(self):

        localctx = JSONParser.PropnameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_propname)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 75
            _la = self._input.LA(1)
            if not(_la==JSONParser.IDENTIFIER or _la==JSONParser.STRING):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ArrayContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def value(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(JSONParser.ValueContext)
            else:
                return self.getTypedRuleContext(JSONParser.ValueContext,i)


        def getRuleIndex(self):
            return JSONParser.RULE_array

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterArray" ):
                listener.enterArray(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitArray" ):
                listener.exitArray(self)




    def array(self):

        localctx = JSONParser.ArrayContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_array)
        self._la = 0 # Token type
        try:
            self.state = 90
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,8,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 77
                self.match(JSONParser.T__4)
                self.state = 78
                self.value()
                self.state = 83
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==JSONParser.T__0:
                    self.state = 79
                    self.match(JSONParser.T__0)
                    self.state = 80
                    self.value()
                    self.state = 85
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 86
                self.match(JSONParser.T__8)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 88
                self.match(JSONParser.T__4)
                self.state = 89
                self.match(JSONParser.T__8)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx


    class ValueContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STRING(self):
            return self.getToken(JSONParser.STRING, 0)

        def NUMBER(self):
            return self.getToken(JSONParser.NUMBER, 0)

        def obj(self):
            return self.getTypedRuleContext(JSONParser.ObjContext,0)


        def envelope(self):
            return self.getTypedRuleContext(JSONParser.EnvelopeContext,0)


        def array(self):
            return self.getTypedRuleContext(JSONParser.ArrayContext,0)


        def getRuleIndex(self):
            return JSONParser.RULE_value

        def enterRule(self, listener:ParseTreeListener):
            if hasattr( listener, "enterValue" ):
                listener.enterValue(self)

        def exitRule(self, listener:ParseTreeListener):
            if hasattr( listener, "exitValue" ):
                listener.exitValue(self)




    def value(self):

        localctx = JSONParser.ValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_value)
        try:
            self.state = 100
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,9,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 92
                self.match(JSONParser.STRING)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 93
                self.match(JSONParser.NUMBER)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 94
                self.obj()
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 95
                self.envelope()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 96
                self.array()
                pass

            elif la_ == 6:
                self.enterOuterAlt(localctx, 6)
                self.state = 97
                self.match(JSONParser.T__9)
                pass

            elif la_ == 7:
                self.enterOuterAlt(localctx, 7)
                self.state = 98
                self.match(JSONParser.T__10)
                pass

            elif la_ == 8:
                self.enterOuterAlt(localctx, 8)
                self.state = 99
                self.match(JSONParser.T__11)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx









class AnsonException(BaseException):
    type = "io.odysz.ansons.x.AnsonException"
    excode = 0
    err = ""
    
    def __init__(self, excode, err):
        self.excode = excode
        self.err = err



'''
Created on 25 Oct 2019

@author: odys-z@github.com
'''
import os
import sys
from re import match
from typing import TextIO, Optional, TypeVar

T = TypeVar('T')


class LangExt:
    '''
    classdocs
    '''

    def __init__(self, params):
        '''
        Constructor
        '''

    @staticmethod
    def isblank(s, regex=None):
        if (s == None):
            return True
        if isinstance(s, str):
            if regex == None or s == "":
                return len(s) == 0
            else:
                return match(s, regex)
        return False

    @staticmethod
    def ifnull(a: T, b: T) -> T:
        return b if a is None else a

    @classmethod
    def len(cls, obj):
        return 0 if obj is None else len(obj)

    @staticmethod
    def str(obj: dict | list):
        def quot(v) -> str:
            return f'"{v}"' if type(v) == str else f'"{v.toBlock()}"' if isinstance(v, Anson) else LangExt.str(v)
        from .ansons import Anson
        if type(obj) == dict:
            s = '{'
            for k, v in obj.items():
                # s += f'{"" if len(s) == 1 else ",\n"}"{k}": "{LangExt.str(v)}"'
                s += f'{"" if len(s) == 1 else ",\n"}"{k}": {quot(v)}'
            s += '}'
            return s
        elif type(obj) == list:
            s = '['
            # s += ", ".join(f'"{x}"' if type(x) == str else LangExt.str(x) for x in obj)
            s += ", ".join(quot(x) for x in obj)
            return s + ']'
        elif isinstance(obj, Anson):
            return obj.toBlock()
        else:
            return str(obj)

def log(out: Optional[TextIO], templt: str, *args):
    try:
        print(templt if LangExt.isblank(args) else templt.format(*args), file=out)
    except Exception as e:
        print(templt, args, e)


class Utils:
    def __init__(self, params):
        '''
        Constructor
        '''

    @staticmethod
    def logi(templt, *args):
        log(sys.stdout, templt, *args)

    @staticmethod
    def warn(templt, *args):
        log(sys.stderr, templt, *args)

    @staticmethod
    def get_os():
        """
        :return: Windows | Linux | macOS
        """
        if os.name == 'nt':
            return 'Windows'
        elif os.name == 'posix':
            if sys.platform.startswith('linux') or sys.platform.startswith('freebsd'):
              return 'Linux'
            elif sys.platform.startswith('darwin'):
                return 'macOS'
        return 'Unknown'

    @staticmethod
    def iswindows():
        return Utils.get_os() == 'Windows'



'''
Created on 28 Oct 2019

Data types for test cases.

@author: odys-z@github.com
'''
from src.anson.io.odysz.ansons import Anson


class AnsT1(Anson):
    ver = "1.0" # type: str
    
    m = None # type: AnsM1 



import sys
import unittest

from src.anson.io.odysz.ansons import Anson
from test.io.odysz.jclient import OnError, Clients
from test.io.odysz.semantic.jprotocol import AnsonResp


class AnclientTest(unittest.TestCase):
    def testPing(self):
        Anson.java_src('test')
        err = OnError(lambda c, e, args: print(c, e.format(args), file=sys.stderr) and self.fail(e))

        Clients.servRt = 'http://127.0.0.1:8964/jserv-album'
        resp = Clients.pingLess('Anson.py3/test', err)
        self.assertIsNotNone(resp)

        print(Clients.servRt, '<echo>', resp.toBlock())
        self.assertEqual(type(resp.body[0]), AnsonResp)
        self.assertEqual('ok', resp.code) # TODO MsgCode.ok


if __name__ == '__main__':
    unittest.main()
    t = AnclientTest()
    t.testPing()




'''
Created on 25 Oct 2019

@author: odys-z@github.com
'''

from dataclasses import dataclass
import inspect
from dataclasses import dataclass
from enum import Enum
from src.ansonpy_del.JSONListener import JSONListener
from src.anson.io.odysz.common import LangExt, Utils
from src.ansonpy_del.x import AnsonException
import abc
import decimal
from typing import List
import re
from importlib import import_module


class AnsonFlags():
    parser = True


################################# Anson ##################################
class IJsonable(abc.ABC):
    '''
    Java interface (protocol) can be deserailized into json.
    For python protocol and ABC, see
    http://masnun.rocks/2017/04/15/interfaces-in-python-protocols-and-abcs/
    '''

    @abc.abstractmethod
    def toBlock(self, outstream):
        pass


def writeVal(outstream, v):
    if isinstance(v, str):
        outstream.write("\"")
        outstream.write(v)
        outstream.write("\"")
    else:
        outstream.write(str(v))


@dataclass
class Anson(IJsonable):
    to_del: str = "some vale"
    to_del_int: int = 5

    def toBlock(self, outstream, opts):
        quotK = opts == None or opts.length == 0 or opts[0] == None or opts[0].quotKey();
        if (quotK == True):
            outstream.write("{\"type\": \"");
            outstream.write(self.getClass());
            outstream.write('\"');
        else:
            outstream.write("{type: ");
            outstream.write(self.getClass());

        for (n, v) in self.getFields():
            outstream.write(", ");
            if (quotK == True):
                outstream.write("\"%s\": " % n);
            else:
                outstream.write("%s: " % n);
            writeVal(outstream, v);

        outstream.write("}")
        return "";

    def getClass(self):
        return "io.odysz.ansons.Anson"

    def getFields(self):
        env_dict = []
        for (name, att) in inspect.getmembers(self, lambda attr: not callable(attr)):
            if (not name.startswith("__")):
                env_dict.append((name, att))
        return env_dict


################################# From Json ##############################
class Reflectn(object):
    moduclss = dict()

    ''' Reflection helper
    '''

    @staticmethod
    def isArray(typename):
        return re.match(typename, 'list')

    @staticmethod
    def parseElemType(subTypes: str) -> List[str]:
        ''' Java equivalent:
            private static String[] parseElemType(String subTypes)
        '''
        if (LangExt.isblank(subTypes)):
            return None;
        return subTypes.split("/", 2);

    @staticmethod
    def isAnson(modclss) -> bool:
        ''' Is the class declared as a subclass of Anson?
            This verification also prevent injection attack?
            see discussion about eval():
            https://stackoverflow.com/questions/3451779/how-to-dynamically-create-an-instance-of-a-class-in-python

            This class will also automatically verifyClss user's module.class

            Also DP verify module class name.
            Parameters:
            -----------
            modclss: str
                name of module.path.ClassName
        '''
        modu, clss = modclss.rsplit('.', 1)
        if modclss not in Reflectn.moduclss:
            Reflectn.moduclss.update({modclss: modu})
            return clss, modclss
        else:
            return None, None

    @staticmethod
    def creatByName(class_str: str = 'tutorial.foo.foo.Class1'):
        try:
            module_path, class_name = class_str.rsplit('.', 1)
            module = import_module(module_path)
            return getattr(module, class_name), eval(class_str)()
        except (ImportError, AttributeError) as e:  # @UnusedVariable
            raise ImportError(class_str)

    @staticmethod
    def allSubcls(cls):
        allSubs = []

        for sub in cls.__subclasses__():
            allSubs.append(sub.__name__)
            allSubs.extend(Reflectn.allSubcls(sub))

        return allSubs


# class AnInst:
#     ''' Anson Type
#         Equivolent of java Type
#     '''
#     def __init__(self, ftype):
#         self.ftype = ftype
#
#     def getClass(self):
#         return self.antype

class ParsingCtx():
    ''' Parsing AST node's context, for handling the node's value,
        the element class of parsing stack.

        Attributes
        ---------
        parsingProp: str
            The json prop (object key)
            java: protected String parsingProp;

        parsedVal: object
            The parsed native value

        valType: str
            field value types

        enclosing: object
            enclosing instance currently parsing

        # private HashMap<String, Field> fmap;

        subType: str;
            Annotation's sub types
    '''

    def __init__(self, fmap, enclosing):
        '''
        Parameters
        ----------
        fmap: map
            fields map

        enclosing: object
            enclosing object
        '''
        self.fmap = fmap;
        self.enclosing = enclosing;

    def isInList(self):
        '''
        Returns
        -------
        boolean
            is in a list
        '''
        return isinstance(self.enclosing, list)

    def isInMap(self):
        return isinstance(self.enclosing, dict);

    def elemType(self, tn):
        ''' Set list element's type

            Parameters
            ----------
                tn : list of str
                    type name list
            Returns
            -------
                ParsingCtx: this
        '''
        self.valType = None if tn == None or tn.length <= 0 else tn[0];
        self.subTypes = None if tn == None or tn.length <= 1 else tn[1];

        #         if (not LangExt.isblank(self.valType)):
        #             # change / replace array type
        #             # e.g. lang.String[] to [Llang.String;
        #             if (self.valType.matches(".*\\[\\]$")):
        #                 self.valType = "[L" + self.valType.replaceAll("\\[\\]$", ";");
        #                 self.valType = self.valType.replaceFirst("^\\[L\\[", "[[");
        return self;

    #     def elemType(self):
    #         ''' Get type annotation
    #             Returns
    #             -------
    #                 value type, {@link AnsonField#valType()} annotation if possible
    #         '''
    #         return self.valType;

    def subTypes(self):
        return self.subTypes;


class AnsonListener(JSONListener):
    def __init__(self):
        self.stack = []

    factorys = None
    ''' static
        private static HashMap<Class<?>, JsonableFactory> factorys;
    '''

    # envetype = None;
    # ''' Envelope Type Name '''

    # stack = []
    # ''' Parsing Node Stack '''

    def toparent(self, anInst):
        if (self.stack.size() <= 1 or LangExt.isblank(anInst, "None")):
            # no enclosing, no parent
            return None;

        # trace back, guess with type for children could be in array or map
        # ParsingCtx
        p = self.stack.get(1);
        i = 2;
        while (p != None):
            if (anInst == p.enclosing.getClass()):
                return p.enclosing;
            p = self.stack.get(i);
            i = i + 1;
        return None;

    def push(self, enclosingClazz, elemType=None):
        ''' Push parsing node (a envelope, map, list).
            private void push(Class<?> enclosingClazz, String[] elemType)
        Parameters
        ----------
        enclosingClazz : str
            new parsing IJsonable object's class

        elemType : [type] Not used in python?
            annotation of enclosing list/array. 0: main type, 1: sub-types
            This parameter can't be None if is pushing a list node.

        Raises
        ------
        ReflectiveOperationException
        SecurityException
        AnsonException
        '''
        if (Reflectn.isArray(enclosingClazz)):
            # HashMap<String, Field>
            fmap = map();
            # ParsingCtx
            newCtx = ParsingCtx(fmap, list());
            self.stack.insert(0, newCtx.elemType(elemType));
        else:
            fmap = {};  # HashMap<String, Field>

            if (JSONListener.isArray(enclosingClazz)):
                enclosing = list();
                self.stack.insert(0, ParsingCtx(fmap, enclosing).elemType(elemType));
            else:
                #                 Constructor<?> ctor = None;
                #                 try:
                #                     ctor = enclosingClazz.getConstructor();
                #                 except NoSuchMethodException as e:
                #                     throw new AnsonException(0, "To make json can be parsed to %s, the class must has a default constructor(0 parameter)\n"
                #                             + "Also, inner class must be static."
                #                             + "getConstructor error: %s %s",
                #                             enclosingClazz.getName(), e.getClass().getName(), e.getMessage());
                #                 if (ctor != None && IJsonable.class.isAssignableFrom(enclosingClazz)):
                if (Reflectn.isAnson(enclosingClazz)):
                    # fmap = mergeFields(enclosingClazz, fmap); # map merging is only needed by typed object
                    fmap = {}
                    try:
                        # IJsonable
                        # enclosing = newInstance();
                        enclosing = Anson();
                        self.stack.insert(0, ParsingCtx(fmap, enclosing));
                    except Exception as e:
                        raise AnsonException(0, "Failed to create instance of IJsonable with\nconstructor: %s\n"
                                             + "class: %s\nerror: %s\nmessage: %s\n"
                                             + "Make sure the object can be created with the constructor.",
                                             enclosingClazz, enclosingClazz.getName(), e.getClass().getName(),
                                             e.getMessage());
                else:
                    enclosing = {};
                    # ParsingCtx
                    top = ParsingCtx(fmap, enclosing);
                    self.stack.insert(0, top);

    def pop(self):
        ''' private ParsingCtx pop() {
        Returns
        -------
            ParsingCtx
        '''
        top = self.stack.popleft();
        return top;

    #     def enterJson(self, ctx):
    #         # print("Hello: %s" % ctx.envelope()[0].type_pair().TYPE())
    #         self.an = Anson()

    def exitObj(self):
        top = self.pop();
        top().parsedVal = top.enclosing;
        top.enclosing = None;

    def enterObj(self, ctx):
        top = self.top();
        try:
            fmap = top.fmap if top != None else None;
            if (fmap == None or not fmap.containsKey(top.parsingProp)):
                # In a list, found object, if not type specified with annotation, must failed.
                # But this is confusing to user. Set some report here.
                if (top.isInList() or top.isInMap()):
                    Utils.warn("Type in list or map is complicate, but no annotation for type info can be found. "
                               + "field type: %s\njson: %s\n"
                               + "Example: @AnsonField(valType=\"io.your.type\")\n"
                               + "Anson instances don't need annotation, but objects in json array without type-pair can also trigger this error report.",
                               top.enclosing.getClass(), ctx.getText());
                raise AnsonException(0, "Obj type not found. property: %s", top.parsingProp);

            # Class<?>
            ft = fmap.get(top.parsingProp).getType();
            # if (Map.class.isAssignableFrom(ft)):
            if (isinstance(fmap.get(top.parsingProp), dict)):
                # entering a map
                self.push(ft, None);
                # append annotation
                # Field
                f = top.fmap.get(top.parsingProp);
                #                 # AnsonField
                #                 a = None if f == None else f.getAnnotation(AnsonField.class);
                #                 String anno = a == None ? None : a.valType();
                #
                #                 if (anno != None):
                #                     String[] tn = parseElemType(anno);
                #                     top().elemType(tn);
                top().elemType("object")
            else:
                # entering an envelope
                # push(fmap.get(top.parsingProp).getType());
                self.push(ft, None);
        except (AnsonException) as e:
            e.printStackTrace();

    def parsedEnvelope(self) -> IJsonable:
        if self.stack:
            return self.stack[0].enclosing;
        else:
            raise AnsonException(0, "No envelope is avaliable.");

    def enterEnvelope(self, ctx) -> None:
        ''' Java equivalent:
            public void enterEnvelope(EnvelopeContext)
        '''
        if (self.stack == None):
            self.stack = [];
        self.envetype = None;

    def exitEnvelope(self, ctx) -> None:
        ''' Java equivalent:
            public void exitEnvelope(EnvelopeContext)
        '''
        super().exitEnvelope(ctx);
        if (len(self.stack) > 1):
            top = self.pop();  # ParsingCtx
            top().parsedVal = top.enclosing;
        # else keep last one (root) as return value

    def enterType_pair(self, ctx) -> None:
        ''' Semantics of entering a type pair is found and parsingVal an IJsonable object.<br>
            This is always happening on entering an object.
            The logic opposite to this is exit object.

            Java equivalent:
            public void enterType_pair(Type_pairContext)
            @see gen.antlr.json.JSONBaseListener#enterType_pair(gen.antlr.json.JSONParser.Type_pairContext)
        '''
        if (self.envetype != None):
            # ignore this type specification, keep consist with java type
            return;

        strType = ctx.qualifiedName().STRING();  # TerminalNode
        txt = ctx.qualifiedName().getText();  # String
        envetype = AnsonListener.getStringValRaw(
            strType if isinstance(strType, str) else strType.getText(),
            txt);

        try:
            self.push(envetype);
        except (AnsonException) as e:
            e.printStackTrace();

    def enterPair(self, ctx) -> None:
        ''' Java equivalent:
            public void enterPair(PairContext ctx)
        '''
        super.enterPair(ctx);
        top = self.top();  # ParsingCtx
        top.parsingProp = self.getProp(ctx);
        top.parsedVal = None;

    #     private static String[] parseListElemType(Field f) throws AnsonException {
    #     @staticmethod
    #     def parseListElemType(f) -> list[str]:
    #         # for more information, see
    #         # https://stackoverflow.com/questions/1868333/how-can-i-determine-the-type-of-a-generic-field-in-java
    #
    #         # Type
    #         typ = f.getGenericType();
    #         if (isinstance(typ, ParameterizedType)):
    #             # ParameterizedType
    #             pType = typ;
    #
    #             # String[]
    #             ptypess = pType.getActualTypeArguments()[0].getTypeName().split("<", 2);
    #             if (ptypess.length > 1):
    #                 ptypess[1] = ptypess[1].replaceFirst(">$", "");
    #                 ptypess[1] = ptypess[1].replaceFirst("^L", "");
    #             # figure out array element class
    #             else :
    #                 # Type
    #                 argType = pType.getActualTypeArguments()[0];
    #                 if (not isinstance(argType, TypeVariable) and not isinstance(argType, WildcardType)):
    #                     # Class<? extends Object>
    #                     eleClzz = argType;
    #                     if (eleClzz.isArray()):
    #                         ptypess = list[ptypess[0], eleClzz.getComponentType().getName()];
    #                 # else nothing can do here for a type parameter, e.g. "T"
    #                 elif (AnsonFlags.parser):
    #                         Utils.warn("[AnsonFlags.parser] Element type <%s> for %s is a type parameter (%s) - ignored",
    #                             pType.getActualTypeArguments()[0],
    #                             f.getName(),
    #                             pType.getActualTypeArguments()[0].getClass());
    #             return ptypess;
    #         elif (f.getType().isArray()):
    #             # complex array may also has annotation
    #             # AnsonField a = None if f == None else f.getAnnotation(AnsonField.class);
    #             a = "object";
    #             # String
    #             tn = None if a == None else a.valType();
    #             # String[]
    #             valss = JSONListener.parseElemType(tn);
    #
    #             eleType = f.getType().getComponentType().getTypeName();
    #             if (valss != None and not eleType.equals(valss[0])):
    #                 Utils.warn("[JSONAnsonListener#parseListElemType()]: Field %s is not annotated correctly.\n"
    #                         + "field parameter type: %s, annotated element type: %s, annotated sub-type: %s",
    #                         f.getName(), eleType, valss[0], valss[1]);
    #
    #             if (valss != None and valss.length > 1):
    #                 return list[eleType, valss[1]];
    #             else:
    #                 return list[eleType];
    #         else :
    #             # not a parameterized, not an array, try annotation
    #             # AnsonField a = f == None ? None : f.getAnnotation(AnsonField.class);
    #             a = "object"
    #             tn = None if a == None else a.valType();
    #             return JSONListener.parseElemType(tn);

    #     /**Parse property name, tolerate enclosing quotes presenting or not.
    #      * @param ctx
    #      * @return
    #      */
    #     private static String getProp(PairContext ctx) {
    @staticmethod
    def getProp(ctx) -> str:
        # TerminalNode
        p = ctx.propname().IDENTIFIER();
        return ctx.propname().STRING().getText().replaceAll("(^\\s*\"\\s*)|(\\s*\"\\s*$)",
                                                            "") if p == None else p.getText();

    @staticmethod
    def getStringVal(ctx) -> str:
        ''' Java equivalent:
            private static String getStringVal(PairContext ctx)

            Convert json value : STRING | NUMBER | 'true' | 'false' | 'None' to java.lang.String.<br>
            java: Can't handle NUMBER | obj | array.

            parameters:
            -----------
                ctx: PairContext
                    antlr parsing context
            returns:
            -----------
                value in string
        '''

        # TerminalNode
        stri = ctx.value().STRING();
        # String
        txt = ctx.value().getText();
        return AnsonListener.getStringValRaw(stri, txt);

    @staticmethod
    def getStringValRaw(stri, rawTxt) -> str:
        ''' Get string value.
            If stri is null, use rawTxt.
            Java equivalent:
            private static String getStringVal(TerminalNode str, String rawTxt) {
        '''
        if (stri == None):
            try:
                if (LangExt.isblank(rawTxt)):
                    return None;
                else:
                    if ("None".equals(rawTxt)):
                        return None;
            except Exception as e:
                {}  # @UnusedVariable
            return rawTxt;
        else:
            # stri.getText().replaceAll("(^\\s*\")|(\"\\s*$)", "")
            return re.sub(r'(^\s*")|("\s*$)', '', stri)

    @staticmethod
    def figureJsonVal(ctx) -> object:
        '''grammar:<pre>value
        : STRING
        | NUMBER
        | obj        // all array's obj value can't parsed as Anson, taken as HashMap
        | envelope
        | array
        | 'true'
        | 'false'
        | 'None'
        ;</pre>
        Parameters:
        -----------
            ctx: ValueContext
        Returns
        -------
            java simple value (STRING, NUMBER, 'true', 'false', None)

        java equivalent:
        private static Object figureJsonVal(ValueContext ctx
        '''
        txt = ctx.getText();
        if (txt == None):
            return None;
        elif (ctx.NUMBER() != None):
            try:
                return int(txt);
            except Exception as e:  # @UnusedVariable
                try:
                    return float(txt);
                except Exception as e1:  # @UnusedVariable
                    return decimal(txt);
        elif (ctx.STRING() != None):
            return AnsonListener.getStringVal(ctx.STRING(), txt);
        elif (txt != None and txt.toLowerCase().equals("true")):
            return True;
        elif (txt != None and txt.toLowerCase().equals("flase")):
            return False;
        return None;

    #     @Override
    #     public void enterArray(ArrayContext ctx) {
    def enterArray(self, ctx) -> None:
        try:
            # ParsingCtx
            top = self.top();

            # if in a list or a map, parse top's sub-type as the new node's value type
            if (top.isInList() or top.isInMap()):
                # pushing ArrayList.class because entering array, isInMap() == true means needing to figure out value type
                # String[]
                # tn = JSONListener.parseElemType(top.subTypes());
                # ctx:        [{type:io.odysz.ansons.AnsT2,s:4},{type:io.odysz.ansons.AnsT1,ver:"x"}]
                # subtype:    io.odysz.ansons.Anson
                # tn :        [io.odysz.ansons.Anson]
                # push(ArrayList.class, tn);
                self.push(list, 'object');
            # if field available, parse field's value type as the new node's value type
            else:
                # Class<?>
                ft = top.fmap.get(top.parsingProp).getType();
                # Field
                f = top.fmap.get(top.parsingProp);
                # AnsT3 { ArrayList<Anson[]> ms; }
                # ctx: [[{type:io.odysz.ansons.AnsT2,s:4},{type:io.odysz.ansons.AnsT1,ver:"x"}]]
                # [0]: io.odysz.ansons.Anson[],
                # [1]: io.odysz.ansons.Anson
                # String[]
                tn = AnsonListener.parseListElemType(f);
                self.push(ft, tn);

            # now top is the enclosing list, it's component type is elem-type

        except (AnsonException) as e:
            e.printStackTrace();

    #     @Override
    #     public void exitArray(ArrayContext ctx) {
    #         if (!top().isInList())
    #             throw new NullPointerException("existing not from an eclosing list. txt:\n" + ctx.getText());
    #
    #         ParsingCtx top = pop();
    #         List<?> arr = (List<?>) top.enclosing;
    #
    #         top = top();
    #         top.parsedVal = arr;
    #
    #         // figure the type if possible - convert to array
    #         String et = top.elemType();
    #         if (!LangExt.isblank(et, "\\?.*")) // TODO debug: where did this type comes from?
    #             try {
    #                 Class<?> arrClzz = Class.forName(et);
    #                 if (arrClzz.isArray())
    #                     top.parsedVal = toPrimitiveArray(arr, arrClzz);
    #             } catch (AnsonException | IllegalArgumentException | ClassNotFoundException e) {
    #                 Utils.warn("Trying convert array to annotated type failed.\ntype: %s\njson: %s\nerror: %s",
    #                         et, ctx.getText(), e.getMessage());
    #             }
    #         // No annotation, for 2d list, parsed value is still a list.
    #         // If enclosed element of array is also an array, it can not been handled here
    #         // Because there is no clue for sub array's type if annotation is empty
    #     }
    #
    #     def toPrimitiveArray(list, arrType):
    #         '''
    #         private static <P> P toPrimitiveArray(List<?> list, Class<P> arrType) throws AnsonException {
    #          * Unboxes a List in to a primitive array.
    #          * reference:
    #          * https://stackoverflow.com/questions/25149412/how-to-convert-listt-to-array-t-for-primitive-types-using-generic-method
    #          *
    #          * @param  list      the List to convert to a primitive array
    #          * @param  arrType the primitive array type to convert to
    #          * @param  <P>       the primitive array type to convert to
    #          * @return an array of P with the elements of the specified List
    #          * @throws AnsonException list element class doesn't equal array element type - not enough annotation?
    #          * @throws NullPointerException
    #          *         if either of the arguments are None, or if any of the elements
    #          *         of the List are None
    #          * @throws IllegalArgumentException
    #          *         if the specified Class does not represent an array type, if
    #          *         the component type of the specified Class is not a primitive
    #          *         type, or if the elements of the specified List can not be
    #          *         stored in an array of type P
    #         '''
    #         if (!arrType.isArray()):
    #             throw new IllegalArgumentException(arrType.toString());
    #
    #         if (list == None):
    #             return None;
    #
    #         Class<?> eleType = arrType.getComponentType();
    #
    #         P array = arrType.cast(Array.newInstance(eleType, list.size()));
    #
    #         for (int i = 0; i < list.size(); i++):
    #             Object lstItem = list.get(i);
    #             if (lstItem == None)
    #                 continue;
    #
    #             # this guess is error prone, let's tell user why. May be more annotation is needed
    #             if (!eleType.isAssignableFrom(lstItem.getClass()))
    #                 throw new AnsonException(1, "Set element (v: %s, type %s) to array of type of \"%s[]\" failed.\n"
    #                         + "Array element's type not annotated?",
    #                         lstItem, lstItem.getClass(), eleType);
    #
    #             Array.set(array, i, list.get(i));
    #
    #         return array;
    #
    #     public void exitValue(ValueContext ctx) {
    #         '''
    #         grammar:<pre>value
    #         : STRING
    #         | NUMBER
    #         | obj        // all array's obj value can't parsed as Anson, taken as HashMap - TODO doc: known issue
    #         | envelope
    #         | array
    #         | 'true'
    #         | 'false'
    #         | 'None'
    #         ;</pre>
    #          * @see gen.antlr.json.JSONBaseListener#exitValue(gen.antlr.json.JSONParser.ValueContext)
    #         '''
    #         ParsingCtx top = top();
    #         if (top.isInList() || top.isInMap()) {
    #             # if in a map, parsingProp is the map key,
    #             # element type can only been handled with a guess,
    #             # or according to annotation
    #             # String txt = ctx.getText();
    #             if (top.isInList()) {
    #                 List<?> enclosLst = (List<?>) top.enclosing;
    #                 # for List, ft is not None
    #                 if (top.parsedVal == None) {
    #                     # simple value like String or number
    #                     ((List<Object>)enclosLst).add(figureJsonVal(ctx));
    #                 }
    #                 else {
    #                     # try figure out is element also an array if enclosing object is an array
    #                     # e.g. convert elements of List<String> to String[]
    #                     # FIXME issue: if the first element is 0 length, it will failed to convert the array
    #                     Class<?> parsedClzz = top.parsedVal.getClass();
    #                     if (List.class.isAssignableFrom(parsedClzz)) {
    #                         if (LangExt.isblank(top.elemType(), "\\?.*")) {
    #                             // change list to array
    #                             List<?> lst = (List<?>)top.parsedVal;
    #                             if (lst != None && lst.size() > 0) {
    #                                 // search first non-None element's type
    #                                 Class<? extends Object> eleClz = None;
    #                                 int ix = 0;
    #                                 while (ix < lst.size() && lst.get(ix) == None)
    #                                     ix++;
    #                                 if (ix < lst.size())
    #                                     eleClz = lst.get(ix).getClass();
    #
    #                                 if (eleClz != None) {
    #                                     try {
    #                                         ((List<Object>)enclosLst).add(toPrimitiveArray(lst,
    #                                                 Array.newInstance(eleClz, 0).getClass()));
    #                                     } catch (AnsonException e) {
    #                                         Utils.warn("Trying convert array to annotated type failed.\nenclosing: %s\njson: %s\nerror: %s",
    #                                             top.enclosing, ctx.getText(), e.getMessage());
    #                                     }
    #
    #                                     # remember elem type for later None element
    #                                     top.elemType(new String[] {eleClz.getName()});
    #                                 }
    #                                 # all elements are None, ignore the list is the only way
    #                             }
    #                             else
    #                                 # FIXME this will broken when first element's length is 0.
    #                                 ((List<Object>)enclosLst).add(lst.toArray());
    #                         }
    #                         # branch: with annotation or type name already figured out from 1st element
    #                         else {
    #                             try {
    #                                 List<?> parsedLst = (List<?>)top.parsedVal;
    #                                 String eleType = top.elemType();
    #                                 Class<?> eleClz = Class.forName(eleType);
    #                                 if (eleClz.isAssignableFrom(parsedClzz)) {
    #                                     # annotated element can be this branch
    #                                     ((List<Object>)enclosLst).add(parsedLst);
    #                                 }
    #                                 else {
    #                                     # type is figured out from the previous element,
    #                                     # needing conversion to array
    #                                     #
    #                                     # Bug: object value can't been set into string array
    #                                     # lst.getClass().getTypeName() = java.lang.ArrayList
    #                                     # ["val",88.91669145042222]
    #
    #
    #                                     # Test case:    AnsT3 { ArrayList<Anson[]> ms; }
    #                                     # ctx:         [{type:io.odysz.ansons.AnsT2,s:4},{type:io.odysz.ansons.AnsT1,ver:"x"}]
    #                                     # parsedLst:    [{type: io.odysz.ansons.AnsT2, s: 4, m: None}, {type: io.odysz.ansons.AnsT1, ver: "x", m: None}]
    #                                     # parsedClzz:    java.util.ArrayList
    #                                     # eleType:        [Lio.odysz.ansons.Anson;
    #                                     # eleClz:        class [Lio.odysz.ansons.Anson;
    #                                     # action - change parsedLst to array, add to enclosLst
    #                                     ((List<Object>)enclosLst).add(toPrimitiveArray(parsedLst,
    #                                                     Array.newInstance(eleClz, 0).getClass()));
    #                             except Exception as e:
    #                                 Utils.warn(envelopName());
    #                                 Utils.warn(ctx.getText());
    #                                 e.printStackTrace();
    #                     else:
    #                         ((List<Object>)enclosLst).add(top.parsedVal);
    #                 top.parsedVal = None;
    #             else if (top.isInMap()):
    #                 # parsed Value can already got when exit array
    #                 if (top.parsedVal == None)
    #                     top.parsedVal = getStringVal(ctx.STRING(), ctx.getText());
    #
    #     public void exitPair(PairContext ctx) {
    #         super.exitPair(ctx);
    #         if (AnsonFlags.parser) {
    #             Utils.logi("[AnsonFlags.parser] Property-name: %s", ctx.getChild(0).getText());
    #             Utils.logi("[AnsonFlags.parser] Property-value: %s", ctx.getChild(2).getText());
    #         }
    #
    #         try {
    #             // String fn = getProp(ctx);
    #             ParsingCtx top = top();
    #             String fn = top.parsingProp;
    #
    #             // map's pairs also exits here - map helper
    #             if (top.isInMap()) {
    #                 ((HashMap<String, Object>)top.enclosing).put(top.parsingProp, top.parsedVal);
    #                 top.parsedVal = None;
    #                 top.parsingProp = None;
    #                 return;
    #             }
    #             # not map ...
    #
    #             Object enclosing = top().enclosing;
    #             Field f = top.fmap.get(fn);
    #             if (f == None)
    #                 throw new AnsonException(0, "Field ignored: field: %s, value: %s", fn, ctx.getText());
    #
    #             f.setAccessible(true);
    #             AnsonField af = f.getAnnotation(AnsonField.class);
    #             if (af != None && af.ignoreFrom()) {
    #                 if (AnsonFlags.parser)
    #                     Utils.logi("[AnsonFlags.parser] %s ignored", fn);
    #                 return;
    #             }
    #             else if (af != None && af.ref() == AnsonField.enclosing) {
    #                 Object parent = toparent(f.getType());
    #                 if (parent == None)
    #                     Utils.warn("parent %s is ignored: reference is None", fn);
    #
    #                 f.set(enclosing, parent);
    #                 return;
    #             }
    #
    #             Class<?> ft = f.getType();
    #
    #             if (ft == String.class) {
    #                 String v = getStringVal(ctx);
    #                 f.set(enclosing, v);
    #             }
    #             else if (ft.isPrimitive()) {
    #                 # construct primitive value
    #                 v = ctx.getChild(2).getText();
    #                 setPrimitive((IJsonable) enclosing, f, v);
    #             else if (ft.isEnum()):
    #                 String v = getStringVal(ctx);
    #                 if (!LangExt.isblank(v)):
    #                     f.set(enclosing, Enum.valueOf((Class<Enum>) ft, v));
    #             else if (ft.isArray())
    #                 f.set(enclosing, toPrimitiveArray((List<?>)top.parsedVal, ft));
    #             else if (List.class.isAssignableFrom(ft)
    #                     or AbstractCollection.class.isAssignableFrom(ft)
    #                     or Map.class.isAssignableFrom(ft)):
    #                 f.set(enclosing, top.parsedVal);
    #             else if (IJsonable.class.isAssignableFrom(ft)):
    #                 if (Anson.class.isAssignableFrom(ft))
    #                     f.set(enclosing, top.parsedVal);
    #                 else:
    #                     # Subclass of IJsonable must registered
    #                     String v = getStringVal(ctx);
    #                     if (!LangExt.isblank(v, "None"))
    #                         f.set(enclosing, invokeFactory(f, v));
    #             else if (Object.class.isAssignableFrom(ft)):
    #                 Utils.warn("\nDeserializing unsupported type, field: %s, type: %s, enclosing type: %s",
    #                         fn, ft.getName(), enclosing == None ? None : enclosing.getClass().getName());
    #                 String v = ctx.getChild(2).getText();
    #
    #                 if (!LangExt.isblank(v, "None"))
    #                     f.set(enclosing, v);
    #             else throw new AnsonException(0, "sholdn't happen");
    #
    #             # not necessary, top is dropped
    #             top.parsedVal = None;
    #         except (ReflectiveOperationException, RuntimeException) as e:
    #             e.printStackTrace();
    #         except AnsonException e:
    #             Utils.warn(e.getMessage());

    def invokeFactory(self, f, v):
        '''
        private IJsonable invokeFactory(Field f, String v) throws AnsonException {
        '''
        if (self.factorys == None or not self.factorys.containsKey(f.getType())):
            raise AnsonException(0,
                                 "Subclass of IJsonable (%s) must registered.\n - See javadoc of IJsonable.JsonFacotry\n"
                                 + "Or don't declare the field as %1$s, use a subclass of Anson",
                                 f.getType());

        factory = self.factorys.get(f.getType());
        try:
            return factory.fromJson(v);
        except Exception as t:
            raise AnsonException(0,
                                 "Subclass of IJsonable (%s) must registered.\n - See javadoc of IJsonable.JsonFacotry\n"
                                 + "Or don't declare the field as %1$s, use a subclass of Anson",
                                 f.getType(), t.getMessage());

    @staticmethod
    def setPrimitive(self, obj, f, v):
        '''
        private static void setPrimitive(IJsonable obj, Field f, String v)
            throws RuntimeException, ReflectiveOperationException, AnsonException {
        '''

    #         if (f.getType() == int.class || f.getType() == Integer.class)
    #             f.set(obj, Integer.valueOf(v));
    #         else if (f.getType() == float.class || f.getType() == Float.class)
    #             f.set(obj, Float.valueOf(v));
    #         else if (f.getType() == double.class || f.getType() == Double.class)
    #             f.set(obj, Double.valueOf(v));
    #         else if (f.getType() == long.class || f.getType() == Long.class)
    #             f.set(obj, Long.valueOf(v));
    #         else if (f.getType() == short.class || f.getType() == Short.class)
    #             f.set(obj, Short.valueOf(v));
    #         else if (f.getType() == byte.class || f.getType() == Byte.class)
    #             f.set(obj, Byte.valueOf(v));
    #         else
    #             # what's else?
    #             raise AnsonException(0, "Unsupported field type: %s (field %s)",
    #                     f.getType().getName(), f.getName());

    @staticmethod
    def registFactory(jsonable, factory):
        '''
        Parameters
        ---------
        jsonable: Class<?>
        factory: JsonableFactory
        '''
        if (AnsonListener.factorys == None):
            # factorys = new HashMap<Class<?>, JsonableFactory>();
            factorys = {}
        factorys.put(jsonable, factory);


################################## To Json ###############################

class MsgCode(Enum):
    ok = "ok"
    exGeneral = "exGeneral"
    exSemantics = "exSemantics"
    exTransc = "exTransac"


class Port(Enum):
    session = "login.serv"
    r = "r.serv"


class AnsonMsg(Anson):
    # code = MsgCode.ok
    def __init__(self):
        self.port = None

    body = []

    def getClass(self):
        return "io.odysz.ansons.AnsonMsg"  # self.type;


class AnsonBody(Anson):
    pass


class AnsonReq(AnsonBody):
    def __init__(self):
        self.a = None


class AnsonResp(AnsonBody):
    def __init__(self):
        self.a = None



from dataclasses import dataclass

from src.anson.io.odysz.ansons import Anson


@dataclass
class Synode(Anson):
    org: str
    synid: str
    mac: str
    domain: str
    nyq: int
    syn_uid: str

    def __init__(self):
        super().__init__()


@dataclass()
class SyncUser(Anson):
    userId: str
    userName: str
    pswd: str
    iv: str
    domain: str
    org: str

    def __init__(self, userId=None, userName=None, pswd=None):
        super().__init__()
        self.userId = userId
        self.userName = userName
        self.pswd = pswd





import unittest
from datetime import datetime

from src.anson.io.odysz.common import LangExt
from test.io.oz.syn import SyncUser


class LangExtTest(unittest.TestCase):
    def testStr(self):
        obj = {'a': 1, 'b': "2"}

        dt = datetime.now()
        lst = [1, 3, 'a', 5.5, dt, str(dt)]

        self.assertEqual('{"a": 1,\n"b": "2"}', LangExt.str(obj))
        self.assertEqual(f'[1, 3, "a", 5.5, {str(dt)}, "{str(dt)}"]', LangExt.str(lst))
        self.assertEqual('1', LangExt.str(1))

        usr = SyncUser(userId='1', userName='ody', pswd='8964')
        self.assertEqual('''{
  "type": "io.odysz.semantic.syn.SyncUser",
  "userId": "1",
  "userName": "ody",
  "pswd": "8964"
}''', LangExt.str(usr))

        usr = {'a': 1, 'b': usr}
        self.assertEqual('''{"a": 1,
"b": "{
  "type": "io.odysz.semantic.syn.SyncUser",
  "userId": "1",
  "userName": "ody",
  "pswd": "8964"
}"}''', LangExt.str(usr))

        self.assertEqual('''[2, {"a": 1,
"b": "{
  "type": "io.odysz.semantic.syn.SyncUser",
  "userId": "1",
  "userName": "ody",
  "pswd": "8964"
}"}]''', LangExt.str([2, usr]))


if __name__ == '__main__':
    unittest.main()
    t = LangExtTest()
    t.testStr()




from typing import get_type_hints

class Node:
    def __init__(self, value: int, next_node: 'Node' = None):
        self.value = value
        self.next_node = next_node

    def get_next(self) -> 'Node':
        return self.next_node

# Get type hints for the __init__ method
hints = get_type_hints(Node.__init__)
print(hints['next_node'])  # Outputs: <class '__main__.Node'>

node = Node(1)  # next_node is None by default
print(type(node.next_node))  # Outputs: <class 'NoneType'>

node2 = Node(1, node)  # Now next_node is another Node
print('node2.next_node', type(node2.next_node))


from collections.abc import Callable
from sys import stderr
from typing import Protocol, Any

import requests

from test.io.odysz import SessionInf
from test.io.odysz.semantic.jprotocol import MsgCode, Port, AnsonMsg, AnsonBody, AnsonResp, AnsonHeader
from test.io.odysz.semantic.jserv.echo import EchoReq, A


class OnError(Protocol):
    err : Callable = None
    def __call__(self, code: MsgCode, msg: str, *args: str) -> None:
        return self.err(code, msg, args)

    def __init__(self, on_err: Callable[[MsgCode, str, ...], None]):
        self.err = on_err


class Clients:
    """
    Java stub
    """

    servRt = None

    @staticmethod
    def pingLess(funcUri: str, errCtx: OnError=None):
        req = EchoReq()
        req.a = A.echo

        client = InsecureClient(Clients.servRt)
        jmsg = client.userReq(funcUri, Port.echo, req)

        resp = client.commit(jmsg, errCtx)

        return resp


class SessionClient:
    myservRt: str
    ssInf: SessionInf
    header: AnsonHeader

    def __init__(self, jserv: str, ssInf: SessionInf):
        self.proxies = {
            "http": None,
            "https": None,
        }
        '''
        https://stackoverflow.com/a/40470853/7362888
        '''

        self.myservRt = jserv
        self.ssInf = ssInf
        self.header = None

    def Header(self):
        if self.header is None:
            self.header = AnsonHeader(ssid=self.ssInf.ssid, uid=self.ssInf.uid, token=self.ssInf.ssToken)
        return self.header

    def commit(self, req: AnsonMsg, err: OnError) -> AnsonResp | None:
        try:
            print(f'{self.myservRt}/{req.port.value}')
            print(req.toBlock(False))
            resp = requests.post(f'{self.myservRt}/{req.port.value}', proxies=self.proxies, data=req.toBlock(False))
            if resp.status_code == 200:
                data = resp.json()  # If the response is JSON
                return AnsonResp.from_envelope(data)
            else:
                print(f"Error: {resp.status_code}", file=stderr)
                res = f'{resp.status_code}\n{self.myservRt}\n{'' if req is None else req.toBlock()}'
                err.err(MsgCode.exIo, res, resp.text)
                return None
        except Exception as e:
            if err is not None:
                err.err(MsgCode.exIo, e.message if hasattr(e, 'message') else str(e), None)
            else:
                raise e

    def userReq(self, uri: str, port: Port, bodyItem: AnsonBody, *act: Any) -> AnsonMsg:
        bodyItem.uri = uri
        if act is not None and len(act) > 0:
            self.Header().act = act

        return AnsonMsg(port).Header(self.Header()).Body(bodyItem)


class InsecureClient(SessionClient):

    def __init__(self, servRt: str):
        super().__init__(servRt, SessionInf('uid', 'session less'))




# This Python file uses the following encoding: utf-8
import os
from dataclasses import dataclass
from typing import overload, Optional

from src.anson.io.odysz.ansons import Anson
from src.anson.io.odysz.common import LangExt

jserv_sep = ' '
synode_sep = ':'


class PortfolioException(Exception):
    """
    As upto Portfolio 0.7, there is no such equivalent in Java.
    date
    Thu 20 Feb 2025 11:52:06 AM AWST
    mvn dependency:tree
    [INFO] io.github.odys-z:jserv-album:jar:0.7.0
    [INFO] +- io.github.odys-z:docsync.jserv:jar:0.2.2-SNAPSHOT:compile
    [INFO] |  +- io.github.odys-z:semantic.DA:jar:1.5.18-SNAPSHOT:compile (version selected from constraint [1.5.18-SNAPSHOT,2.0.0-SNAPSHOT))
    [INFO] |  |  +- io.github.odys-z:semantics.transact:jar:1.5.58:compile (version selected from constraint [1.5.58,))
    [INFO] |  |  |  \\- io.github.odys-z:antson:jar:0.9.114:compile (version selected from constraint [0.9.111,))
    [INFO] |  +- io.github.odys-z:anclient.java:jar:0.5.16:compile (version selected from constraint [0.5.16,))
    [INFO] |  \\- io.github.odys-z:synodict-jclient:jar:0.1.6:compile (version selected from constraint [0.1.6,))
    [INFO] +- io.github.odys-z:syndoc-lib:jar:0.5.18-SNAPSHOT:compile
    [INFO] |  \\- io.github.odys-z:semantic.jserv:jar:1.5.16-SNAPSHOT:compile (version selected from constraint [1.5.16-SNAPSHOT,2.0.0-SNAPSHOT))
    [INFO] +- io.github.odys-z:albumtier:jar:0.5.0-SNAPSHOT:test
    """
    msg: str
    cause: object

    def __init__(self, msg: str, *args: object):
        super().__init__(args)
        self.msg = msg
        self.cause = None if args is None or len(args) == 0 else args[0]


@dataclass
class AppSettings(Anson):
    envars: dict
    rootkey: str    # | None # test 3.12
    installkey: Optional[str] # test 3.9

    volume: str
    vol_name: str
    port: int
    jservs: dict

    def __init__(self):
        super().__init__()
        self.envars = {}

    @overload
    def Volume(self):
        """
        return self.volume
        """

    @overload
    def Volume(self, v: str):
        """
        set volume
        return self
        """

    def Volume(self, v: str = None):
        if v is None:
            return self.volume
        else:
            self.volume = os.path.normpath(v).replace("\\", "/")
            return self

    @overload
    def Jservs(self, jservs: dict):
        ...
        # self.jservs = parseJservs(jservs)
        # return self

    @overload
    def Jservs(self) -> str:
        ...

    def Jservs(self, urldict: dict = None):
        '''
        :param urldict:
            E.g. {x: 'http://127.0.0.1:8964/jserv-album'}
        :return: self when setting, jservs lines, [['x', 'http://127.0.0.1:8964/jserv-album']], when getting.
        '''
        if urldict is None:
            return [[k, self.jservs[k]] for k in self.jservs]
        else:
            self.jservs = urldict
            return self

    def jservLines(self):
        return [':\t'.join([k, self.jservs[k]]) for k in self.jservs]

    # def movekey(self):
    #     if LangExt.len(self.rootkey) == 0 and LangExt.len(self.installkey) > 0:
    #         self.rootkey, self.installkey = self.installkey, None



import json
from dataclasses import dataclass
from pathlib import Path

from src.anson.io.odysz.ansons import Anson
from test.io.odysz.semantic.syn import Synode, SyncUser


@dataclass
class SynOrg(Anson):
    meta: str

    orgId: str
    orgName: str
    # edu | org | com | ...
    orgType: str
    ''' This is a tree table. '''
    parent: str
    sort: str
    fullpath: str
    market: str
    # web server url, configured in dictionary like: $WEB-ROOT:8888
    webroot: str
    # The home page url (landing page)
    homepage: str
    # The default resources collection, usually a group / tree of documents.
    album0: str

    def __init__(self):
        super().__init__()


@dataclass
class SynodeConfig(Anson):
    synid: str
    domain: str

    admin: str

    sysconn: str
    synconn: str

    org: SynOrg
    ''' Market, organization or so? '''

    syncIns: float
    '''
     * Synchronization interval, initially, in secends.
     * No worker thread started if less or equals 0.
    '''

    peers: list[Synode]

    # mode: SynodeMode
    https: bool

    def __init__(self):
        super().__init__()
        self.https = False


@dataclass()
class AnRegistry(Anson):
    config: SynodeConfig
    synusers: list[SyncUser]

    def __init__(self):
        super().__init__()
        self.config = None # SynodeConfig()
        self.synusers = []

    @staticmethod
    def load(path) -> 'AnRegistry':
        if Path(path).is_file():
            with open(path, 'r') as file:
                # registry = json.load(file)
                # self.config = SynodeConfig.from_obj(registry['config'])
                obj = json.load(file)
                obj['__type__'] = AnRegistry().__type__
                return Anson.from_envelope(obj)
        else:
            raise FileNotFoundError(f"File doesn't exist: {path}")

    @classmethod
    def find_synode(cls, synodes: list[Synode], id):
        if synodes is not None:
            for peer in synodes:
                if peer.synid == id:
                    return peer
        return None


def loadYellowPages():
    path = ""
    registry = AnRegistry().load(path)
    return registry
